@@ -1,5 +1,5 @@
 /* hdparm.c - Command line interface to get/set hard disk parameters */
-/*          - by Mark S. Lord (c) 1994-1998 -- freely distributable */
+/*          - by Mark S. Lord (C) 1994-2000 -- freely distributable */
 
 #include <unistd.h>
 #include <string.h>
@@ -19,8 +19,13 @@
 #include <linux/fs.h>
 #include <linux/major.h>
 
-#define VERSION "v3.6"
+#define VERSION "v3.9"
 
+#undef	 DO_FLUSHCACHE		/* under construction: force cache flush on -W0 */
+
+#ifndef CDROM_SELECT_SPEED	/* already defined in 2.3.xx kernels and above */
+#define CDROM_SELECT_SPEED       0x5322
+#endif
 /*
  * For kernels prior to 1.3.61, HDIO_SET_32BIT is HDIO_SET_CHIPSET
  */
@@ -40,7 +45,7 @@
 char *progname;
 static struct hd_driveid id;
 static int verbose = 0, get_identity = 0, get_geom = 0, noisy = 1, quiet = 0;
-static flagcount = 0, do_flush = 0, is_scsi_hd = 0, is_xt_hd = 0;
+static int flagcount = 0, do_flush = 0, is_scsi_hd = 0, is_xt_hd = 0;
 static int do_ctimings, do_timings = 0, old_identity = 0, timing_MB = TIMING_MB;
 
 static unsigned long set_readahead= 0, get_readahead= 0, readahead= 0;     
@@ -76,9 +81,22 @@ static unsigned long set_standbynow = 0, get_standbynow = 0;
 static unsigned long set_sleepnow   = 0, get_sleepnow   = 0;
 static unsigned long get_powermode  = 0;
 #endif
+#ifdef CDROM_SELECT_SPEED
+static unsigned long set_cdromspeed = 0, cdromspeed = 0;
+#endif /* CDROM_SELECT_SPEED */
 #ifdef HDIO_SET_32BIT
 static int get_IDentity = 0;
 #endif
+#ifdef HDIO_UNREGISTER_HWIF
+static int	unregister_hwif = 0;
+static int	hwif = 0;
+#endif
+#ifdef HDIO_SCAN_HWIF
+static int	scan_hwif = 0;
+static int	hwif_data = 0;
+static int	hwif_ctrl = 0;
+static int	hwif_irq = 0;
+#endif
 
 char *cfg_str[] =
 {	"", " HardSect", " SoftSect", " NotMFM", " HdSw>15uSec", " SpinMotCtl",
@@ -87,18 +105,22 @@ char *cfg_str[] =
 };
 
 char *SlowMedFast[]	= {"slow", "medium", "fast", "eide"};
-char *BuffType[]	= {"?", "1Sect", "DualPort", "DualPortCache"};
+char *BuffType[]	= {"unknown", "1Sect", "DualPort", "DualPortCache"};
 
 #define YN(b)	(((b)==0)?"no":"yes")
 
 static void dmpstr (char *prefix, unsigned int i, char *s[], unsigned int maxi)
 {
-	printf("%s%d(%s)", prefix, i, (i > maxi) ? "?" : s[i]);
+	if (i > maxi)
+		printf("%s%d", prefix, i);
+	else
+		printf("%s%s", prefix, s[i]);
 }
 
 static void dump_identity (struct hd_driveid *id)
 {
 	int i;
+	char pmodes[64] = {0,}, dmodes[128]={0,};
 
 	printf("\n Model=%.40s, FwRev=%.8s, SerialNo=%.20s", 
 		id->model, id->fw_rev, id->serial_no);
@@ -122,12 +144,12 @@ static void dump_identity (struct hd_driveid *id)
 		else
 			printf("off");
 	}
-	printf("\n DblWordIO=%s", YN(id->dword_io&1));
-	printf(", OldPIO=%d", id->tPIO);
-	printf(", DMA=%s", YN(id->capability&1));
-	if (id->capability&1)
-		printf(", OldDMA=%d", id->tDMA);
 	putchar('\n');
+	if (id->tPIO <= 5) {
+		strcat(pmodes, "pio0 ");
+		if (id->tPIO >= 1) strcat(pmodes, "pio1 ");
+		if (id->tPIO >= 2) strcat(pmodes, "pio2 ");
+	}
 	if (!(id->field_valid&1))
 		printf(" (maybe):");
 	printf(" CurCHS=%d/%d/%d, CurSects=%d",
@@ -135,77 +157,68 @@ static void dump_identity (struct hd_driveid *id)
 	printf(", LBA=%s", YN(id->capability&2));
 	if (id->capability&2)
  		printf(", LBAsects=%d", id->lba_capacity);
+
 	if (id->capability&1) {
-		if ((id->field_valid&2) || (id->dma_1word | id->dma_mword))
-			printf("\n ");
-		if (id->field_valid&2) {
-			printf("tDMA={min:%d,rec:%d}",
-			 id->eide_dma_min, id->eide_dma_time);
-		}
 		if (id->dma_1word | id->dma_mword) {
-			if (id->field_valid&2)
-				printf(", ");
-			printf("DMA modes: ");
-			if (id->dma_1word & 0x100)	printf("*");
-			if (id->dma_1word & 1)		printf("sword0 ");
-			if (id->dma_1word & 0x200)	printf("*");
-			if (id->dma_1word & 2)		printf("sword1 ");
-			if (id->dma_1word & 0x400)	printf("*");
-			if (id->dma_1word & 4)		printf("sword2 ");
-			if (id->dma_1word & 0xf800)	printf("*");
-			if (id->dma_1word & 0xf8)	printf("sword? ");
-			if (id->dma_mword & 0x100)	printf("*");
-			if (id->dma_mword & 1)		printf("mword0 ");
-			if (id->dma_mword & 0x200)	printf("*");
-			if (id->dma_mword & 2)		printf("mword1 ");
-			if (id->dma_mword & 0x400)	printf("*");
-			if (id->dma_mword & 4)		printf("mword2 ");
-			if (id->dma_mword & 0xf800)	printf("*");
-			if (id->dma_mword & 0xf8)	printf("mword? ");
+			if (id->dma_1word & 0x100)	strcat(dmodes,"*");
+			if (id->dma_1word & 1)		strcat(dmodes,"sdma0 ");
+			if (id->dma_1word & 0x200)	strcat(dmodes,"*");
+			if (id->dma_1word & 2)		strcat(dmodes,"sdma1 ");
+			if (id->dma_1word & 0x400)	strcat(dmodes,"*");
+			if (id->dma_1word & 4)		strcat(dmodes,"sdma2 ");
+			if (id->dma_1word & 0xf800)	strcat(dmodes,"*");
+			if (id->dma_1word & 0xf8)	strcat(dmodes,"sdma? ");
+			if (id->dma_mword & 0x100)	strcat(dmodes,"*");
+			if (id->dma_mword & 1)		strcat(dmodes,"mdma0 ");
+			if (id->dma_mword & 0x200)	strcat(dmodes,"*");
+			if (id->dma_mword & 2)		strcat(dmodes,"mdma1 ");
+			if (id->dma_mword & 0x400)	strcat(dmodes,"*");
+			if (id->dma_mword & 4)		strcat(dmodes,"mdma2 ");
+			if (id->dma_mword & 0xf800)	strcat(dmodes,"*");
+			if (id->dma_mword & 0xf8)	strcat(dmodes,"mdma? ");
 		}
 	}
+	printf("\n IORDY=");
+	if (id->capability&8)
+		printf((id->capability&4) ? "on/off" : "yes");
+	else
+		printf("no");
 	if ((id->capability&8) || (id->field_valid&2)) {
-		printf("\n IORDY=");
-		if (id->capability&8)
-			printf((id->capability&4) ? "on/off" : "yes");
-		else
-			printf("no");
 		if (id->field_valid&2) {
-			printf(", tPIO={min:%d,w/IORDY:%d}, PIO modes: ",
-			 id->eide_pio, id->eide_pio_iordy);
-			if (id->eide_pio_modes & 1)	printf("mode3 ");
-			if (id->eide_pio_modes & 2)	printf("mode4 ");
-			if (id->eide_pio_modes &~3)	printf("mode? ");
+			printf(", tPIO={min:%d,w/IORDY:%d}", id->eide_pio, id->eide_pio_iordy);
+			if (id->eide_pio_modes & 1)	strcat(pmodes, "pio3 ");
+			if (id->eide_pio_modes & 2)	strcat(pmodes, "pio4 ");
+			if (id->eide_pio_modes &~3)	strcat(pmodes, "pio? ");
 		}
+#ifdef HDIO_OBSOLETE_IDENTITY
 		if (id->field_valid&4) {
-			printf("\n UDMA modes: ");
-#ifndef HDIO_OBSOLETE_IDENTITY
-			printf(" (not available: rebuild hdparm with a newer kernel");
-#else
 			if (old_identity) {
-				printf(" (not available from this kernel)");
+				strcat(dmodes," udma??");
 			} else {
-				if (id->dma_ultra & 0x100)	printf("*");
-				if (id->dma_ultra & 0x001)	printf("mode0 ");
-				if (id->dma_ultra & 0x200)	printf("*");
-				if (id->dma_ultra & 0x002)	printf("mode1 ");
-				if (id->dma_ultra & 0x400)	printf("*");
-				if (id->dma_ultra & 0x004)	printf("mode2 ");
-				if (id->dma_ultra & 0x800)	printf("*");
-				if (id->dma_ultra & 0x008)	printf("mode3 ");
-				if (id->dma_ultra & 0x1000)	printf("*");
-				if (id->dma_ultra & 0x0010)	printf("mode4 ");
-				if (id->dma_ultra & 0x2000)	printf("*");
-				if (id->dma_ultra & 0x0020)	printf("mode5 ");
-				if (id->dma_ultra & 0x4000)	printf("*");
-				if (id->dma_ultra & 0x0040)	printf("mode6 ");
-				if (id->dma_ultra & 0x8000)	printf("*");
-				if (id->dma_ultra & 0x0080)	printf("mode7 ");
+				if (id->dma_ultra & 0x100)	strcat(dmodes,"*");
+				if (id->dma_ultra & 0x001)	strcat(dmodes,"udma0 ");
+				if (id->dma_ultra & 0x200)	strcat(dmodes,"*");
+				if (id->dma_ultra & 0x002)	strcat(dmodes,"udma1 ");
+				if (id->dma_ultra & 0x400)	strcat(dmodes,"*");
+				if (id->dma_ultra & 0x004)	strcat(dmodes,"udma2 ");
+				if (id->dma_ultra & 0x800)	strcat(dmodes,"*");
+				if (id->dma_ultra & 0x008)	strcat(dmodes,"udma3 ");
+				if (id->dma_ultra & 0x1000)	strcat(dmodes,"*");
+				if (id->dma_ultra & 0x0010)	strcat(dmodes,"udma4 ");
+				if (id->dma_ultra & 0x2000)	strcat(dmodes,"*");
+				if (id->dma_ultra & 0x0020)	strcat(dmodes,"udma5 ");
+				if (id->dma_ultra & 0x4000)	strcat(dmodes,"*");
+				if (id->dma_ultra & 0x0040)	strcat(dmodes,"udma6 ");
+				if (id->dma_ultra & 0x8000)	strcat(dmodes,"*");
+				if (id->dma_ultra & 0x0080)	strcat(dmodes,"udma7 ");
 			}
 #endif
 		}
 	}
-	printf("\n\n");
+	if ((id->capability&1) && (id->field_valid&2))
+		printf(", tDMA={min:%d,rec:%d}", id->eide_dma_min, id->eide_dma_time);
+	printf("\n PIO modes: %s", pmodes);
+	printf("\n DMA modes: %s\n", dmodes);
 }
 
 void flush_buffer_cache (int fd)
@@ -218,7 +231,7 @@ void flush_buffer_cache (int fd)
 		sleep(1);
 	} else {
 		if (ioctl(fd, HDIO_DRIVE_CMD, NULL))	/* await completion */
-			perror("HDIO_DRIVE_CMD failed");
+			perror("HDIO_DRIVE_CMD(null) failed");
 	}
 #endif
 }
@@ -567,12 +580,24 @@ void process_dev (char *devname)
 #ifdef IDE5_MAJOR
 		 && major(stat_buf.st_rdev) != IDE5_MAJOR
 #endif
-	                                                      ) {
+#ifdef IDE6_MAJOR
+		 && major(stat_buf.st_rdev) != IDE6_MAJOR
+#endif
+#ifdef IDE7_MAJOR
+		 && major(stat_buf.st_rdev) != IDE7_MAJOR
+#endif
+#ifdef IDE8_MAJOR
+		 && major(stat_buf.st_rdev) != IDE8_MAJOR
+#endif
+#ifdef IDE9_MAJOR
+		 && major(stat_buf.st_rdev) != IDE9_MAJOR
+#endif
+	) {
 		fprintf(stderr,"%s not supported by hdparm\n",devname);
 		exit(EINVAL);
 	}
 
-	fd = open (devname, O_RDONLY);
+	fd = open (devname, O_RDONLY|O_NONBLOCK);
 	if (fd < 0) {
 		perror(devname);
 		exit(errno);
@@ -585,6 +610,26 @@ void process_dev (char *devname)
 		if (ioctl(fd, BLKRASET, readahead)) 
 			perror(" BLKRASET failed");
 	}
+#ifdef HDIO_UNREGISTER_HWIF
+	if (unregister_hwif) {
+		no_scsi();
+		printf(" attempting to unregister hwif#%d\n", hwif);
+		if (ioctl(fd, HDIO_UNREGISTER_HWIF, hwif)) 
+			perror(" HDIO_UNREGISTER_HWIF failed");
+	}
+#endif
+#ifdef HDIO_SCAN_HWIF
+	if (scan_hwif) {
+		int	args[3];
+		no_scsi();
+		printf(" attempting to scan hwif (0x%x, 0x%x, %d)\n", hwif_data, hwif_ctrl, hwif_irq);
+		args[0] = hwif_data;
+		args[1] = hwif_ctrl;
+		args[2] = hwif_irq;
+		if (ioctl(fd, HDIO_SCAN_HWIF, args)) 
+			perror(" HDIO_SCAN_HWIF failed");
+	}
+#endif
 #ifdef HDIO_SET_PIO_MODE
 	if (set_piomode) {
 		no_scsi();
@@ -684,7 +729,7 @@ void process_dev (char *devname)
 			on_off(doorlock);
 		}
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
-			perror(" HDIO_DRIVE_CMD failed");
+			perror(" HDIO_DRIVE_CMD(doorlock) failed");
 	}
 	if (set_dkeep) {
 		/* lock/unlock the drive's "feature" settings */
@@ -697,7 +742,7 @@ void process_dev (char *devname)
 		}
 		args[2] = dkeep ? 0x66 : 0xcc;
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
-			perror(" HDIO_DRIVE_CMD failed");
+			perror(" HDIO_DRIVE_CMD(keepsettings) failed");
 	}
 	if (set_defects) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,0x04,0};
@@ -706,7 +751,7 @@ void process_dev (char *devname)
 		if (get_prefetch)
 			printf(" setting drive defect-mgmt to %ld\n", defects);
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
-			perror(" HDIO_DRIVE_CMD failed");
+			perror(" HDIO_DRIVE_CMD(defectmgmt) failed");
 	}
 	if (set_prefetch) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,0xab,0};
@@ -716,7 +761,7 @@ void process_dev (char *devname)
 		if (get_prefetch)
 			printf(" setting drive prefetch to %ld\n", prefetch);
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
-			perror(" HDIO_DRIVE_CMD failed");
+			perror(" HDIO_DRIVE_CMD(setprefetch) failed");
 	}
 	if (set_xfermode) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,3,0};
@@ -728,7 +773,7 @@ void process_dev (char *devname)
 			interpret_xfermode(xfermode);
 		}
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
-			perror(" HDIO_DRIVE_CMD failed");
+			perror(" HDIO_DRIVE_CMD(setxfermode) failed");
 	}
 	if (set_lookahead) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,0,0};
@@ -740,9 +785,22 @@ void process_dev (char *devname)
 			on_off(lookahead);
 		}
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
-			perror(" HDIO_DRIVE_CMD failed");
+			perror(" HDIO_DRIVE_CMD(setreadahead) failed");
 	}
+#ifdef CDROM_SELECT_SPEED
+	if (set_cdromspeed) {
+		printf ("setting cdrom speed to %ld\n", cdromspeed);
+		if (ioctl (fd, CDROM_SELECT_SPEED, cdromspeed))
+			perror(" CDROM_SELECT_SPEED failed");
+	}
+#endif                                    
 	if (set_wcache) {
+#ifdef DO_FLUSHCACHE
+#ifndef WIN_FLUSHCACHE
+#define WIN_FLUSHCACHE 0xe7
+#endif
+		unsigned char flushcache[4] = {WIN_FLUSHCACHE,0,0,0};
+#endif /* DO_FLUSHCACHE */
 		unsigned char args[4] = {WIN_SETFEATURES,0,0,0};
 		no_scsi();
 		no_xt();
@@ -751,12 +809,24 @@ void process_dev (char *devname)
 			printf(" setting drive write-caching to %ld", wcache);
 			on_off(wcache);
 		}
+#ifdef DO_FLUSHCACHE
+		if (!wcache && ioctl(fd, HDIO_DRIVE_CMD, &flushcache))
+			perror (" HDIO_DRIVE_CMD(flushcache) failed");
+#endif /* DO_FLUSHCACHE */
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
-			perror(" HDIO_DRIVE_CMD failed");
+			perror(" HDIO_DRIVE_CMD(setcache) failed");
+#ifdef DO_FLUSHCACHE
+		if (!wcache && ioctl(fd, HDIO_DRIVE_CMD, &flushcache))
+			perror (" HDIO_DRIVE_CMD(flushcache) failed");
+#endif /* DO_FLUSHCACHE */
 	}
 	if (set_standbynow) {
+#ifndef WIN_STANDBYNOW1
 #define WIN_STANDBYNOW1 0xE0
+#endif
+#ifndef WIN_STANDBYNOW2
 #define WIN_STANDBYNOW2 0x94
+#endif
 		unsigned char args1[4] = {WIN_STANDBYNOW1,0,0,0};
 		unsigned char args2[4] = {WIN_STANDBYNOW2,0,0,0};
 		no_scsi();
@@ -764,11 +834,15 @@ void process_dev (char *devname)
 			printf(" issuing standby command\n");
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args1)
 		 && ioctl(fd, HDIO_DRIVE_CMD, &args2))
-			perror(" HDIO_DRIVE_CMD failed");
+			perror(" HDIO_DRIVE_CMD(standby) failed");
 	}
 	if (set_sleepnow) {
+#ifndef WIN_SLEEPNOW1
 #define WIN_SLEEPNOW1 0xE6
+#endif
+#ifndef WIN_SLEEPNOW2
 #define WIN_SLEEPNOW2 0x99
+#endif
 		unsigned char args1[4] = {WIN_SLEEPNOW1,0,0,0};
 		unsigned char args2[4] = {WIN_SLEEPNOW2,0,0,0};
 		no_scsi();
@@ -776,7 +850,7 @@ void process_dev (char *devname)
 			printf(" issuing sleep command\n");
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args1)
 		 && ioctl(fd, HDIO_DRIVE_CMD, &args2))
-			perror(" HDIO_DRIVE_CMD failed");
+			perror(" HDIO_DRIVE_CMD(sleep) failed");
 	}
 	if (set_seagate) {
 		unsigned char args[4] = {0xfb,0,0,0};
@@ -785,7 +859,7 @@ void process_dev (char *devname)
 		if (get_seagate)
 			printf(" disabling Seagate auto powersaving mode\n");
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
-			perror(" HDIO_DRIVE_CMD failed");
+			perror(" HDIO_DRIVE_CMD(seagatepwrsave) failed");
 	}
 	if (set_standby) {
 		unsigned char args[4] = {WIN_SETIDLE1,standby,0,0};
@@ -796,7 +870,7 @@ void process_dev (char *devname)
 			interpret_standby(standby);
 		}
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
-			perror(" HDIO_DRIVE_CMD failed");
+			perror(" HDIO_DRIVE_CMD(setidle1) failed");
 	}
 #else	/* HDIO_DRIVE_CMD */
 	if (force_operation) {
@@ -813,9 +887,10 @@ void process_dev (char *devname)
 	if ((verbose && !is_scsi_hd && !is_xt_hd) || get_mult || get_identity) {
 		no_scsi();
 		multcount = -1;
-		if (ioctl(fd, HDIO_GET_MULTCOUNT, &multcount))
-			perror(" HDIO_GET_MULTCOUNT failed");
-		else if (verbose | get_mult) {
+		if (ioctl(fd, HDIO_GET_MULTCOUNT, &multcount)) {
+			if ((verbose && !is_xt_hd) || get_mult)
+				perror(" HDIO_GET_MULTCOUNT failed");
+		} else if (verbose | get_mult) {
 			printf(" multcount    = %2ld", multcount);
 			on_off(multcount);
 		}
@@ -914,8 +989,12 @@ void process_dev (char *devname)
 	}
 #ifdef HDIO_DRIVE_CMD
 	if (get_powermode) {
+#ifndef WIN_CHECKPOWERMODE1
 #define WIN_CHECKPOWERMODE1 0xE5
+#endif
+#ifndef WIN_CHECKPOWERMODE2
 #define WIN_CHECKPOWERMODE2 0x98
+#endif
 		unsigned char args[4] = {WIN_CHECKPOWERMODE1,0,0,0};
 		const char *state;
 		no_scsi();
@@ -966,7 +1045,7 @@ void process_dev (char *devname)
 				old_identity = 0;
 				dump_identity((struct hd_driveid *)&args[4]);
 			} else
-				perror(" HDIO_DRIVE_CMD failed");
+				perror(" HDIO_DRIVE_CMD(identify) failed");
 		}
 	}
 #endif
@@ -984,62 +1063,71 @@ void usage_error (void)
 	fprintf(stderr,"\n%s - get/set hard disk parameters - version %s\n\n", progname, VERSION);
 	fprintf(stderr,"Usage:  %s  [options] [device] ..\n\n", progname);
 	fprintf(stderr,"Options:\n"
-	" -a    get/set fs readahead\n"
+	" -a   get/set fs readahead\n"
 #ifdef HDIO_DRIVE_CMD
-	" -A  * set drive read-lookahead flag (0/1)\n"
+	" -A   set drive read-lookahead flag (0/1)\n"
 #endif
 #ifdef HDIO_SET_32BIT
-	" -c  * get/set IDE 32-bit IO setting\n"
+	" -c   get/set IDE 32-bit IO setting\n"
 #endif
 #ifdef HDIO_DRIVE_CMD
-	" -C  * check IDE power mode status\n"
+	" -C   check IDE power mode status\n"
 #endif
 #ifdef HDIO_SET_DMA
-	" -d  * get/set using_dma flag\n"
+	" -d   get/set using_dma flag\n"
 #endif
-	" -D  * enable/disable drive defect-mgmt\n"
-	" -f    flush buffer cache for device on exit\n"
-	" -g    display drive geometry\n"
-	" -h    display terse usage information\n"
-	" -i  * display drive identification\n"
+	" -D   enable/disable drive defect-mgmt\n"
+#ifdef CDROM_SELECT_SPEED
+	" -E   set cd-rom drive speed\n"
+#endif CDROM_SELECT_SPEED
+	" -f   flush buffer cache for device on exit\n"
+	" -g   display drive geometry\n"
+	" -h   display terse usage information\n"
+	" -i   display drive identification\n"
 #ifdef HDIO_SET_32BIT
-	" -I  * read drive identification directly from drive\n"
+	" -I   read drive identification directly from drive\n"
 #endif
 #ifdef HDIO_GET_KEEPSETTINGS
-	" -k  * get/set keep_settings_over_reset flag (0/1)\n"
+	" -k   get/set keep_settings_over_reset flag (0/1)\n"
 #endif
 #ifdef HDIO_DRIVE_CMD
-	" -K  * set drive keep_features_over_reset flag (0/1)\n"
-	" -L  * set drive doorlock (0/1) (removable harddisks only)\n"
+	" -K   set drive keep_features_over_reset flag (0/1)\n"
+	" -L   set drive doorlock (0/1) (removable harddisks only)\n"
 #endif
-	" -m  * get/set multiple sector count\n"
+	" -m   get/set multiple sector count\n"
 #ifdef HDIO_SET_NOWERR 
-	" -n  * get/set ignore-write-errors flag (0/1)\n"
+	" -n   get/set ignore-write-errors flag (0/1)\n"
 #endif
 #ifdef HDIO_SET_PIO_MODE
-	" -p  * set PIO mode on IDE interface chipset (0,1,2,3,4,...)\n"
+	" -p   set PIO mode on IDE interface chipset (0,1,2,3,4,...)\n"
 #endif
 #ifdef HDIO_DRIVE_CMD
-	" -P  * set drive prefetch count\n"
+	" -P   set drive prefetch count\n"
+#endif
+	" -q   change next setting quietly\n"
+	" -r   get/set readonly flag (DANGEROUS to set)\n"
+#ifdef HDIO_SCAN_HWIF
+	" -R   register an IDE interface (DANGEROUS)\n"
 #endif
-	" -q    change next setting quietly\n"
-	" -r    get/set readonly flag (DANGEROUS to set)\n"
 #ifdef HDIO_DRIVE_CMD
-	" -S  * set standby (spindown) timeout\n"
+	" -S   set standby (spindown) timeout\n"
 #endif
-	" -t    perform device read timings\n"
-	" -T    perform cache read timings\n"
-	" -u  * get/set unmaskirq flag (0/1)\n"
-	" -v    default; same as -acdgkmnru (-gr for SCSI, -adgr for XT)\n"
-	" -V    display program version and exit immediately\n"
+	" -t   perform device read timings\n"
+	" -T   perform cache read timings\n"
+	" -u   get/set unmaskirq flag (0/1)\n"
+#ifdef HDIO_UNREGISTER_HWIF
+	" -U   un-register an IDE interface (DANGEROUS)\n"
+#endif
+	" -v   default; same as -acdgkmnru (-gr for SCSI, -adgr for XT)\n"
+	" -V   display program version and exit immediately\n"
 #ifdef HDIO_DRIVE_CMD
-	" -W  * set drive write-caching flag (0/1) (DANGEROUS)\n"
-	" -X  * set IDE xfer mode (DANGEROUS)\n"
-	" -y  * put IDE drive in standby mode\n"
-	" -Y  * put IDE drive to sleep\n"
-	" -Z  * disable Seagate auto-powersaving mode\n"
+	" -W   set drive write-caching flag (0/1) (DANGEROUS)\n"
+	" -X   set IDE xfer mode (DANGEROUS)\n"
+	" -y   put IDE drive in standby mode\n"
+	" -Y   put IDE drive to sleep\n"
+	" -Z   disable Seagate auto-powersaving mode\n"
 #endif
-	"        * = (E)IDE drives only\n\n");
+	);
 	exit(1);
 }
 
@@ -1181,7 +1269,14 @@ int main(int argc, char **argv)
 						if (!set_defects)
 							fprintf(stderr, "-D: missing value\n");
 						break;
-
+#ifdef CDROM_SELECT_SPEED
+					case 'E':
+						set_cdromspeed = 1;
+						GET_NUMBER(set_cdromspeed,cdromspeed);
+						if (!set_cdromspeed)
+							fprintf (stderr, "-s: missing value\n");
+						break;
+#endif /* CDROM_SELECT_SPEED */
 					case 'P':
 						get_prefetch = noisy;
 						noisy = 1;
@@ -1281,6 +1376,56 @@ int main(int argc, char **argv)
 						}
 						break;
 #endif /* HDIO_GET_KEEPSETTINGS */
+#ifdef HDIO_UNREGISTER_HWIF
+					case 'U':
+						if (!*p && argc && isdigit(**argv))
+							p = *argv++, --argc;
+
+						if(! p) {
+							fprintf(stderr, "expected hwif_nr\n");
+							exit(1);
+						}
+
+						sscanf(p++, "%i", &hwif);
+
+						unregister_hwif = 1;
+						break;
+#endif /* HDIO_UNREGISTER_HWIF */
+#ifdef HDIO_SCAN_HWIF
+					case 'R':
+						if (!*p && argc && isdigit(**argv))
+							p = *argv++, --argc;
+
+						if(! p) {
+							fprintf(stderr, "expected hwif_data\n");
+							exit(1);
+						}
+
+						sscanf(p++, "%i", &hwif_data);
+
+						if (argc && isdigit(**argv))
+							p = *argv++, --argc;
+						else {
+							fprintf(stderr, "expected hwif_ctrl\n");
+							exit(1);
+						}
+
+						sscanf(p, "%i", &hwif_ctrl);
+
+						if (argc && isdigit(**argv))
+							p = *argv++, --argc;
+						else {
+							fprintf(stderr, "expected hwif_irq\n");
+							exit(1);
+						}
+
+						sscanf(p, "%i", &hwif_irq);
+
+						*p = '\0';
+
+						scan_hwif = 1;
+						break;
+#endif /* HDIO_SCAN_HWIF */
 					case 'a':
 						get_readahead = noisy;
 						noisy = 1;
